@@ -11,8 +11,8 @@ use POSIX qw( strftime );
 
 my $loop = IO::Async::Loop->new;
 
-my @digits = map { Tickit::Widget::SegmentDisplay->new( type => '7' ) } 0 .. 5;
-my @colons = map { Tickit::Widget::SegmentDisplay->new( type => ':' ) } 0 .. 1;
+my @digits = map { Tickit::Widget::SegmentDisplay->new( type => '7', use_halfline => 1, thickness => 1 ) } 0 .. 5;
+my @colons = map { Tickit::Widget::SegmentDisplay->new( type => ':', use_halfline => 1, thickness => 1 ) } 0 .. 1;
 
 my $nexttime = time;
 $loop->add( IO::Async::Timer::Absolute->new(
