@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2013-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2013-2015 -- leonerd@leonerd.org.uk
 
 package Tickit::Widget::SegmentDisplay;
 
@@ -13,7 +13,7 @@ use Tickit::Style;
 
 use utf8;
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 use Carp;
 
@@ -70,7 +70,9 @@ use constant WIDGET_PEN_FROM_STYLE => 1;
 
 =cut
 
-=head2 $segmentdisplay = Tickit::Widget::SegmentDisplay->new( %args )
+=head2 new
+
+   $segmentdisplay = Tickit::Widget::SegmentDisplay->new( %args )
 
 Constructs a new C<Tickit::Widget::SegmentDisplay> object.
 
@@ -90,7 +92,7 @@ The type of display. Supported types are:
 
 =item seven
 
-A 7-segment bar display
+A 7-segment bar display. The display can also be blanked with the value " ".
 
 =item seven_dp
 
@@ -103,10 +105,12 @@ A static C<:>
 
 =item symb
 
-A unit or prefix symbol character. The following characters are recognised
+A unit, prefix symbol or other character. The following characters are
+recognised:
 
-  V A W Ω
-  M k m µ
+  V A W Ω F H s
+  G M k m µ n p
+  + - %
 
 Each will be drawn in a style approximately to fit the general LED shape
 display, by drawing lines of erased cells. Note however that some more
@@ -114,6 +118,23 @@ intricate shapes may not be very visible on smaller scales.
 
 =back
 
+=item use_unicode => BOOL
+
+If true, use Unicode block-drawing characters. If false, use only coloured
+erase cells using the background colour.
+
+=item use_halfline => BOOL
+
+If true, vertical resolution of rendered block characters is effectively
+doubled by using half-filled Unicode block-drawing characters. Setting this
+option implies C<use_unicode>.
+
+=item thickness => INT
+
+Gives the number of columns wide and half-lines tall that LED bars will be
+drawn in. Note that unless C<use_halfline> is set, this value ought to be an
+even number. Defaults to 2.
+
 =back
 
 =cut
@@ -142,6 +163,18 @@ sub new
    $self->{reshape_method} = $self->can( "reshape_${method}" );
    $self->{render_method}  = $self->can( "render_${method}_to_rb" );
 
+   my $use_halfline = $args{use_halfline};
+   $self->{use_halfline} = $use_halfline;
+
+   my $use_unicode  = $args{use_unicode};
+
+   $self->{flush_method} = $self->can(
+      $use_halfline ? "flush_halfline" :
+      $use_unicode  ? "flush_unicode"  :
+                      "flush" );
+
+   $self->{thickness} = $args{thickness} // 2;
+
    $self->{value} = $args{value} // "";
 
    $self->on_style_changed_values(
@@ -162,9 +195,11 @@ sub cols  { 4 + 2 }
 
 =cut
 
-=head2 $value = $segmentdisplay->value
+=head2 value
 
-=head2 $segmentdisplay->set_value( $value )
+   $value = $segmentdisplay->value
+
+   $segmentdisplay->set_value( $value )
 
 Return or set the character on display
 
@@ -188,8 +223,8 @@ sub on_style_changed_values
    my $self = shift;
    my %values = @_;
 
-   $self->{lit_pen}   = Tickit::Pen::Immutable->new( bg => $values{lit}[1]   ) if $values{lit};
-   $self->{unlit_pen} = Tickit::Pen::Immutable->new( bg => $values{unlit}[1] ) if $values{unlit};
+   $self->{lit_pen}   = Tickit::Pen::Immutable->new( fg => $values{lit}[1]   ) if $values{lit};
+   $self->{unlit_pen} = Tickit::Pen::Immutable->new( fg => $values{unlit}[1] ) if $values{unlit};
 }
 
 sub reshape
@@ -197,21 +232,145 @@ sub reshape
    my $self = shift;
    my $win = $self->window or return;
 
-   $self->{reshape_method}->( $self, $win->lines, $win->cols, 0, 0 );
+   my $linescale = 1 + !!$self->{use_halfline};
+
+   $self->{reshape_method}->( $self, $win->lines * $linescale, $win->cols, 0, 0 );
 }
 
+use constant {
+   LIT   => 0x01,
+   UNLIT => 0x02,
+};
+
 sub render_to_rb
 {
    my $self = shift;
    my ( $rb, $rect ) = @_;
 
+   my @buff;
+
+   # TODO: sizing?
+   $self->{render_method}->( $self, \@buff );
+
    $rb->eraserect( $rect );
 
-   $self->{render_method}->( $self, $rb, $rect );
+   $self->{flush_method}->( $self, \@buff, $rb, $rect );
+}
+
+sub flush
+{
+   my $self = shift;
+   my ( $buff, $rb, $rect ) = @_;
+
+   my $lit_pen   = Tickit::Pen::Immutable->new( bg => $self->{lit_pen}->getattr( "fg" ) );
+   my $unlit_pen = Tickit::Pen::Immutable->new( bg => $self->{unlit_pen}->getattr( "fg" ) );
+
+   foreach my $line ( $rect->linerange ) {
+      next unless defined( my $cells = $buff->[$line] );
+      foreach my $col ( $rect->left .. $rect->right - 1 ) {
+         my $val = vec( $cells, $col, 2 ) or next;
+         $rb->setpen( $val == LIT ? $lit_pen : $unlit_pen );
+         $rb->erase_at( $line, $col, 1 );
+      }
+   }
+}
+
+use constant {
+   U_FULL  => 0x2588,
+   U_UPPER => 0x2580,
+   U_LOWER => 0x2584,
+};
+
+sub flush_unicode
+{
+   my $self = shift;
+   my ( $buff, $rb, $rect ) = @_;
+
+   my $lit_pen   = $self->{lit_pen};
+   my $unlit_pen = $self->{unlit_pen};
+
+   foreach my $line ( $rect->linerange ) {
+      next unless defined( my $cells = $buff->[$line] );
+      foreach my $col ( $rect->left .. $rect->right - 1 ) {
+         my $val = vec( $cells, $col, 2 ) or next;
+         $rb->setpen( $val == LIT ? $lit_pen : $unlit_pen );
+         $rb->char_at( $line, $col, U_FULL );
+      }
+   }
+}
+
+sub flush_halfline
+{
+   my $self = shift;
+   my ( $buff, $rb, $rect ) = @_;
+
+   my $lit_pen   = $self->{lit_pen};
+   my $unlit_pen = $self->{unlit_pen};
+
+   my $both_pen = Tickit::Pen::Immutable->new(
+      fg => $lit_pen->getattr( 'fg' ),
+      bg => $unlit_pen->getattr( 'fg' ),
+   );
+
+   foreach my $phyline ( $rect->linerange ) {
+      my $hicells = $buff->[$phyline*2];
+      my $locells = $buff->[$phyline*2 + 1];
+
+      next unless defined $hicells or defined $locells;
+
+      $hicells //= "";
+      $locells //= "";
+
+      foreach my $col ( $rect->left .. $rect->right - 1 ) {
+         my $hival = vec( $hicells, $col, 2 );
+         my $loval = vec( $locells, $col, 2 );
+
+         $hival or $loval or next;
+
+         if( $hival == $loval ) {
+            $rb->setpen( ( $hival || $loval ) == LIT ? $lit_pen : $unlit_pen );
+            $rb->char_at( $phyline, $col, U_FULL );
+         }
+         elsif( !$hival or !$loval ) {
+            $rb->setpen( ( $hival || $loval ) == LIT ? $lit_pen : $unlit_pen );
+            $rb->char_at( $phyline, $col, $hival ? U_UPPER : U_LOWER );
+         }
+         else {
+            # Half lit, half unlit
+            $rb->setpen( $both_pen );
+            $rb->char_at( $phyline, $col, $hival == LIT ? U_UPPER : U_LOWER );
+         }
+      }
+   }
+}
+
+sub _fill
+{
+   my $self = shift;
+   my ( $buff, $startline, $endline, $startcol, $endcol, $val ) = @_;
+   $val //= LIT;
+
+   my $thickness = $self->{thickness};
+   my @colrange = ( $startcol .. $endcol + $thickness - 1 );
+
+   $thickness /= 2 unless $self->{use_halfline};
+   my @linerange = ( $startline .. $endline + $thickness - 1 );
+
+   foreach my $line ( @linerange ) {
+      vec( $buff->[$line], $_, 2 ) = $val for @colrange;
+   }
+}
+
+sub _dot
+{
+   my $self = shift;
+   my ( $buff, $line, $col, $val ) = @_;
+   $self->_fill( $buff, $line, $line, $col, $col, $val );
 }
 
 # 7-Segment
 my %segments = (
+   ' ' => "       ",
    0 => "ABCDEF ",
    1 => " BC    ",
    2 => "AB DE G",
@@ -224,15 +383,15 @@ my %segments = (
    9 => "ABCD FG",
 );
 
-sub _pen_for_seg
+sub _val_for_seg
 {
    my $self = shift;
    my ( $segment ) = @_;
 
-   my $segments = $segments{$self->value} or return $self->{unlit_pen};
+   my $segments = $segments{$self->value} or return UNLIT;
 
    my $lit = substr( $segments, ord($segment) - ord("A"), 1 ) ne " ";
-   return $lit ? $self->{lit_pen} : $self->{unlit_pen};
+   return $lit ? LIT : UNLIT;
 }
 
 sub reshape_seven
@@ -240,37 +399,42 @@ sub reshape_seven
    my $self = shift;
    my ( $lines, $cols, $top, $left ) = @_;
 
-   $self->{AGD_col}   = $left + 2;
-   $self->{AGD_width} = $cols - 4;
+   my $thickness = $self->{thickness};
 
-   $self->{FE_col} = $left;
-   $self->{BC_col} = $left + $cols - 2;
+   my $right = $left + $cols - $thickness;
 
-   $self->{A_line} = $top;
-   $self->{G_line} = $top + int( ( $lines - 1 + 0.5 ) / 2 );
-   $self->{D_line} = $top + $lines - 1;
+   $self->{FE_col}       = $left;
+   $self->{AGD_startcol} = $left + $thickness;
+   $self->{AGD_endcol}   = $right - $thickness;
+   $self->{BC_col}       = $right;
+
+   $thickness /= 2 unless $self->{use_halfline};
+
+   my $bottom = $top + $lines - $thickness;
+   my $mid    = int( ( $top + $bottom ) / 2 );
+
+   $self->{A_line}       = $top;
+   $self->{BF_startline} = $top + $thickness;
+   $self->{BF_endline}   = $mid - $thickness;
+   $self->{G_line}       = $mid;
+   $self->{CE_startline} = $mid + $thickness;
+   $self->{CE_endline}   = $bottom - $thickness;
+   $self->{D_line}       = $bottom;
 }
 
 sub render_seven_to_rb
 {
    my $self = shift;
-   my ( $rb ) = @_;
+   my ( $buff ) = @_;
 
-   $rb->erase_at( $self->{A_line}, $self->{AGD_col}, $self->{AGD_width}, $self->_pen_for_seg( "A" ) );
-   $rb->erase_at( $self->{G_line}, $self->{AGD_col}, $self->{AGD_width}, $self->_pen_for_seg( "G" ) );
-   $rb->erase_at( $self->{D_line}, $self->{AGD_col}, $self->{AGD_width}, $self->_pen_for_seg( "D" ) );
+   $self->_fill( $buff, ( $self->{A_line} ) x 2, $self->{AGD_startcol}, $self->{AGD_endcol}, $self->_val_for_seg( "A" ) );
+   $self->_fill( $buff, ( $self->{G_line} ) x 2, $self->{AGD_startcol}, $self->{AGD_endcol}, $self->_val_for_seg( "G" ) );
+   $self->_fill( $buff, ( $self->{D_line} ) x 2, $self->{AGD_startcol}, $self->{AGD_endcol}, $self->_val_for_seg( "D" ) );
 
-   my ( $F_pen, $B_pen ) = ( $self->_pen_for_seg( "F" ), $self->_pen_for_seg( "B" ) );
-   foreach my $line ( $self->{A_line}+1 .. $self->{G_line}-1 ) {
-      $rb->erase_at( $line, $self->{FE_col}, 2, $F_pen );
-      $rb->erase_at( $line, $self->{BC_col}, 2, $B_pen );
-   }
-
-   my ( $E_pen, $C_pen ) = ( $self->_pen_for_seg( "E" ), $self->_pen_for_seg( "C" ) );
-   foreach my $line ( $self->{G_line}+1 .. $self->{D_line}-1 ) {
-      $rb->erase_at( $line, $self->{FE_col}, 2, $E_pen );
-      $rb->erase_at( $line, $self->{BC_col}, 2, $C_pen );
-   }
+   $self->_fill( $buff, $self->{BF_startline}, $self->{BF_endline}, ( $self->{FE_col} ) x 2, $self->_val_for_seg( "F" ) );
+   $self->_fill( $buff, $self->{BF_startline}, $self->{BF_endline}, ( $self->{BC_col} ) x 2, $self->_val_for_seg( "B" ) );
+   $self->_fill( $buff, $self->{CE_startline}, $self->{CE_endline}, ( $self->{FE_col} ) x 2, $self->_val_for_seg( "E" ) );
+   $self->_fill( $buff, $self->{CE_startline}, $self->{CE_endline}, ( $self->{BC_col} ) x 2, $self->_val_for_seg( "C" ) );
 }
 
 # 7-Segment with DP
@@ -288,7 +452,7 @@ sub reshape_seven_dp
 sub render_seven_dp_to_rb
 {
    my $self = shift;
-   my ( $rb ) = @_;
+   my ( $buff ) = @_;
 
    my $value = $self->{value};
    my $dp;
@@ -302,10 +466,9 @@ sub render_seven_dp_to_rb
       $self->{value} = $value;
    }
 
-   $self->render_seven_to_rb( $rb );
+   $self->render_seven_to_rb( $buff );
 
-   my $dp_pen = $dp ? $self->{lit_pen} : $self->{unlit_pen};
-   $rb->erase_at( $self->{DP_line}, $self->{DP_col}, 2, $dp_pen );
+   $self->_dot( $buff, $self->{DP_line}, $self->{DP_col}, $dp ? LIT : UNLIT );
 }
 
 # Static double-dot colon
@@ -326,11 +489,11 @@ sub reshape_colon
 sub render_colon_to_rb
 {
    my $self = shift;
-   my ( $rb ) = @_;
+   my ( $buff ) = @_;
 
    my $col = $self->{colon_col};
-   $rb->erase_at( $self->{A_line}, $col, 2, $self->{lit_pen} );
-   $rb->erase_at( $self->{B_line}, $col, 2, $self->{lit_pen} );
+   $self->_dot( $buff, $self->{A_line}, $col );
+   $self->_dot( $buff, $self->{B_line}, $col );
 }
 
 # Symbol drawing
@@ -345,12 +508,23 @@ my %symbol_strokes = do {
    A => [ [qw( 0,100 50,0 100,100 )], [qw( 20,70 80,70)] ],
    W => [ [qw( 0,0 25,100 50,50 75,100 100,0)] ],
    Ω => [ [qw( 0,100 25,100 25,75 10,60 0,50 0,20 20,0 80,0 100,20 100,50 90,60 75,75 75,100 100,100 ) ] ],
+   F => [ [qw( 0,100 0,0 100,0 )], [qw( 0,50 80,50 )] ],
+   H => [ [qw( 0,0 0,100 )], [qw( 0,50 100,50 )], [qw( 100,0 100,100 )] ],
+   s => [ [qw( 100,50 75,40 25,40 0,50 0,60 25,70 75,70 100,80 100,90 75,100 25,100 0,90 )] ],
 
    # Symbols likely to be used as SI prefixes
+   G => [ [qw( 100,25 65,0 35,0 0,25 0,75 35,100 65,100 100,75 100,50 55,50 )] ],
    M => [ [qw( 0,100 0,0 50,50 100,0 100,100 )] ],
    k => [ [qw( 10,0 10,100 )], [qw( 90,40 10,70 90,100 )] ],
-   m => [ [qw( 0,100 0,50 )], [qw( 10,40 40,40 )], [qw( 50,50 50,100 )], [qw( 60,40 90,40 )], [qw( 90,50 100,100 )] ],
+   m => [ [qw( 0,100 0,50 10,40 40,40 50,50 50,100 )], [qw( 50,50 60,40 90,40 100,50 100,100 )] ],
    µ => [ [qw( 0,100 0,40 )], [qw( 0,80 70,80 80,75 90,60 100,40 )] ],
+   n => [ [qw( 0,100 0,40 )], [qw( 0,50 30,40 70,40 100,50 100,100 )] ],
+   p => [ [qw( 0,100 0,40 )], [qw( 0,55 30,40 70,40 100,55 100,60 70,80 30,80 0,60 )] ],
+
+   # Mathematical symbols
+   '+' => [ [qw( 10,50 90,50 )], [qw( 50,30 50,70 )] ],
+   '-' => [ [qw( 10,50 90,50 )] ],
+   '%' => [ [qw( 10,10 10,30 30,30 30,10 10,10 )], [qw( 20,100 80,00 )], [qw( 70,70 70,90 90,90 90,70 70,70 )] ],
 };
 
 sub reshape_symb
@@ -379,12 +553,10 @@ sub _roundpos
 sub render_symb_to_rb
 {
    my $self = shift;
-   my ( $rb ) = @_;
+   my ( $buff ) = @_;
 
    my $strokes = $symbol_strokes{$self->value} or return;
 
-   $rb->setpen( $self->{lit_pen} );
-
    my $Y_to_line = $self->{Y_to_line};
    my $X_to_col  = $self->{X_to_col};
 
@@ -399,11 +571,11 @@ sub render_symb_to_rb
 
          if( $toL == $atL ) {
             my ( $c, $limC ) = $toC > $atC ? ( $atC, $toC ) : ( $toC, $atC );
-            $rb->erase_at( $atL, $c, $limC - $c + 2 );
+            $self->_fill( $buff, $atL, $atL, $c, $limC );
          }
          elsif( $toC == $atC ) {
             my ( $l, $limL ) = $toL > $atL ? ( $atL, $toL ) : ( $toL, $atL );
-            $rb->erase_at( $_, $atC, 2 ) for $l .. $limL;
+            $self->_fill( $buff, $l, $limL, $atC, $atC );
          }
          else {
             my ( $sL, $eL, $sC, $eC ) = $toL > $atL ? ( $atL, $toL, $atC, $toC )
@@ -423,7 +595,7 @@ sub render_symb_to_rb
                   $c++, $err -= $dL if  $err > $dL;
                   $c--, $err += $dL if -$err > $dL;
 
-                  $rb->erase_at( $l, $c, 2 );
+                  $self->_dot( $buff, $l, $c );
 
                   $err += $dC;
                }
@@ -437,7 +609,7 @@ sub render_symb_to_rb
                   $l++, $err -= $adC if  $err > $adC;
                   $l--, $err += $adC if -$err > $adC;
 
-                  $rb->erase_at( $l, $c, 2 );
+                  $self->_dot( $buff, $l, $c );
 
                   $err += $dL;
                }
